try:
	from openai import OpenAI
except:
	pass
try:
	import google.generativeai as genai
except:
	pass
try:
	from transformers import pipeline
except:
	pass

try:
	from mistralai import Mistral as MistralApi
except:
	pass

try:
	import anthropic
except:
	pass

try:
	import voyageai
except:
	pass

from .uralicApi import lemmatize, get_translation

from .tokenizer import words as tokenize_words

from .string_processing import iso_to_name

from .dictionary_backends import TinyDictionary

from py_markdown_table.markdown_table import markdown_table

import json

from mikatools import pickle_dump, pickle_load

class ModuleNotInstalled(Exception):
	pass

class WordNotInDictionaryException(Exception):
    pass

class NoLemmaException(Exception):
    pass

class NothingToLemmatizeException(Exception):
    pass


class NotImplementedException(Exception):
    pass

def get_llm(llm_name, *args, **kwargs):
	if llm_name == "chatgpt":
		return ChatGPT(*args, **kwargs)
	elif llm_name == "gemini":
		return Gemini(*args, **kwargs)
	elif llm_name == "mistral":
		return Mistral(*args, **kwargs)
	elif llm_name == "claude":
		return Claude(*args, **kwargs)
	elif llm_name == "voyage":
		return Voyage(*args, **kwargs)
	else:
		return HuggingFace(llm_name, *args, **kwargs)
		

class LLM(object):
	"""docstring for LLM"""
	def __init__(self):
		self.cache = False
		self._embed_cache_dict = {}
		self._prompt_cache_dict = {}
		super(LLM, self).__init__()


	def _embed_cache(func):
		def inner(*args, **kwargs):
			self = args[0]
			if self.cache and "_".join(args[1:]) in self._embed_cache_dict:
				return self._embed_cache_dict["_".join(args[1:])]
			else:
				r = func(*args, **kwargs)
				if self.cache:
					self._embed_cache_dict["_".join(args[1:])] = r
				return r
		return inner

	def _prompt_cache(func):
		def inner(*args, **kwargs):
			self = args[0]
			if self.cache and "_".join(args[1:]) in self._prompt_cache_dict:
				return self._prompt_cache_dict["_".join(args[1:])]
			else:
				r = func(*args, **kwargs)
				if self.cache:
					self._prompt_cache_dict["_".join(args[1:])] = r
				return r
		return inner


	@_prompt_cache
	def prompt(self, text):
		return self._prompt(text)

	def _prompt(self, text):
		raise NotImplementedException("LLM does not support prompting")

	@_embed_cache
	def embed(self, text):
		return self._embed(text)

	def _embed(self, text):
		raise NotImplementedException("LLM does not support embeddings")

	@_embed_cache
	def embed_endangered(self, text, lang, dict_lang,backend=TinyDictionary):
		r = []
		for word in tokenize_words(text):
			lemmas = lemmatize(word, lang)
			if len(lemmas) == 0:
				r.append(word)
				continue
			lemma = lemmas[0]
			t = get_translation(lemma, lang, dict_lang,backend=backend)
			if len(t) == 0:
				r.append(lemma)
				continue
			r.append(t[0])
		text = " ".join(r)
		return self.embed(text)

	def save_cache(self, file, *args, **kwargs):
		pickle_dump([self._embed_cache_dict, self._prompt_cache_dict], file, *args, **kwargs )

	def load_cache(self, file, *args, **kwargs):
		self.cache = True
		self._embed_cache_dict, self._prompt_cache_dict = pickle_load(file, *args, **kwargs)


class ChatGPT(LLM):
	"""docstring for ChatGPT"""
	def __init__(self, api_key, model="gpt-4o"):
		super(ChatGPT, self).__init__()
		try:
			self.client = OpenAI(api_key=api_key)
		except:
			raise ModuleNotInstalled("OpenAI Python library is not installed. Run pip install openai. If you do have the library installed, check your API key.")
		self.model = model

	def _prompt(self, prompt, temperature=1):
		chat_completion = self.client.chat.completions.create(
			messages=[
				{
					"role": "user",
					"content": prompt,
				}
			],
			model=self.model,
			temperature=temperature
		)
		return chat_completion.choices[0].message.content

	def _embed(self, text):
		response = self.client.embeddings.create(input=text, model=self.model)
		return response.data[0].embedding

class Gemini(LLM):
	"""docstring for Gemini"""
	def __init__(self, api_key, model="gemini-1.5-flash", task_type="retrieval_document"):
		super(Gemini, self).__init__()
		try:
			genai.configure(api_key=api_key)
		except:
			raise ModuleNotInstalled("Google Python library is not installed. Run pip install google-generativeai. If you do have the library installed, check your API key.")
		self.model = genai.GenerativeModel(model)
		self.model_name = model
		self.task_type = task_type

	def _prompt(self, prompt):
		response = self.model.generate_content(prompt)
		return response.text

	def _embed(self, text):
		result = genai.embed_content(model=self.model_name, content=text, task_type=self.task_type)
		return result['embedding']

class HuggingFace(LLM):
	def __init__(self, model, max_length=1000, device=-1):
		super(HuggingFace, self).__init__()
		self.max_length = max_length
		self.model_name = model
		self.model = None
		self.embedder = None
		self.device = device

	def _prompt(self, prompt):
		if self.model is None:
			self.model = pipeline('text-generation', model = self.model_name, device = self.device)
		r = self.model(prompt, max_length=self.max_length, truncation=True)
		return " ".join([x['generated_text'] for x in r])

	def _embed(self, text):
		if self.embedder is None:
			self.embedder = pipeline('feature-extraction', model=self.model_name,device = self.device)
		r = self.embedder(text, return_tensors="pt")[0].numpy().mean(axis=0)
		return list(r)

class Mistral(LLM):
	def __init__(self, api_key, model="mistral-small-latest"):
		super(Mistral, self).__init__()
		try:
			self.s = MistralApi(api_key=api_key)
		except:
			raise ModuleNotInstalled("Mistral library is not installed. Run pip install mistralai. If you do have the library installed, check your API key.")
		self.model = model

	def _prompt(self, prompt):
		r = self.s.chat.complete(model=self.model, messages=[{"content": prompt,"role": "user",}])
		return r.choices[0].message.content

	def _embed(self, text):
		embeddings_batch_response = self.s.embeddings.create(model=self.model, inputs=[text])
		return embeddings_batch_response.data[0].embedding


class Claude(LLM):
	"""docstring for ChatGPT"""
	def __init__(self, api_key, model="claude-3-5-sonnet-latest", max_length=1024):
		super(Claude, self).__init__()
		try:
			self.client = anthropic.Anthropic(api_key=api_key)
		except:
			raise ModuleNotInstalled("Anthropic Python library is not installed. Run pip install anthropic. If you do have the library installed, check your API key.")
		self.model = model
		self.max_length = max_length

	def _prompt(self, prompt, temperature=1):
		chat_completion = self.client.messages.create(model=self.model,messages=[{"role": "user", "content": prompt}], max_tokens=self.max_length)
		return " ".join([x.text for x in chat_completion.content])


class Voyage(LLM):
	"""docstring for LLM"""
	def __init__(self, api_key, model="voyage-3"):
		super(Voyage, self).__init__()
		try:
			self.vo = voyageai.Client(api_key=api_key)
		except:
			raise ModuleNotInstalled("Voyage Python library is not installed. Run pip install voyageai. If you do have the library installed, check your API key.")
		self.model = model

	def _embed(self, text):
		result = self.vo.embed([text], model=self.model, input_type="document")
		return result.embeddings[0]

def _create_disambiguation_prompt(sentence,lang, dict_lang, raise_exceptions=False, backend=TinyDictionary):
	if isinstance(sentence, str):
		sentence = tokenize_words(sentence)
	prompt = "Your task is to disambiguate a sentence in " + iso_to_name(lang) + ". You will be given the sentence, a table that has all of the words of the sentence in separate rows and a comma separated list of possible lemmas. You will need to pick the correct lemma for each word so that every word will have only one lemma. To help you understand " + iso_to_name(lang) + " you will also get a second table that gives you translations of the words in "+ iso_to_name(dict_lang) + "."
	prompt += "\n\nSentence:\n" + " ".join(sentence)
	prompt += "\n\nTable of lemmas:\n"
	lemma_data = []
	lemmas = []
	max_ambiguity = 1
	for word in sentence:
		w_lemmas = lemmatize(word, lang)
		if raise_exceptions and len(w_lemmas) == 0:
			raise NoLemmaException("No lemmas for word " + word + " in " + lang)
		if len(w_lemmas) > max_ambiguity:
			max_ambiguity = len(w_lemmas)
		lemma_data.append({"Word": word, "Lemmas": ", ".join(w_lemmas)})
		lemmas.extend(w_lemmas)
	if raise_exceptions and max_ambiguity == 1:
		raise NothingToLemmatizeException("The sentence is not ambiguous")
	lemmas = lemmas
	prompt += markdown_table(lemma_data).get_markdown()
	prompt += "\n\n"+ iso_to_name(lang) + " - " + iso_to_name(dict_lang) +" vocabulary:\n"
	vocab_data = []
	for lemma in lemmas:
		trans = get_translation(lemma, lang, dict_lang,backend=backend)
		if raise_exceptions and len(trans) == 0 and lemma not in string.punctuation:
			raise WordNotInDictionaryException("Word " + lemma + " not found in the "+lang + " dictionary for " + dict_lang)
		vocab_data.append({iso_to_name(lang): lemma, iso_to_name(dict_lang):", ".join(trans)})
	prompt += markdown_table(vocab_data).get_markdown()
	prompt += "\n\nPlease write out the steps of your decision process and provide a list of lemmas in JSON format at the very end of your answer. Example: {\"lemmas\": [\"lemma 1\", \"lemma 2\", \"lemma 3\"]}"
	return prompt.replace("```", "")

def disambiguate_sentence(sentence, lang, dict_lang, llm, raise_exceptions=False, backend=TinyDictionary):
	prompt = _create_disambiguation_prompt(sentence, lang, dict_lang, raise_exceptions=raise_exceptions, backend=backend)
	res = llm.prompt(prompt)
	j = res[res.rfind("{\"lemmas\":" ):]
	j = j[:j.find("}")+1]
	try:
		d = json.loads(j)["lemmas"]
	except:
		d = []
	return d, res


		


#encoding: utf-8
import requests, os
import ssl
import copy
import re
import mikatools
from .foma import FomaFSTWrapper
from .string_processing import filter_arabic
from collections.abc import Iterable
import glob
import datetime
import shutil
from .dictionary_backends import TinyDictionary, MongoDictionary
from .neural_fst import NeuralFST

try:
    # For Python 3.0 and later
    from urllib.request import urlopen
    new_python = True
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen
    new_python = False

hfst_backend = "hfst"

try:
	import hfst_dev as hfst
except:
	try:
		import hfst
	except:
		hfst_backend = "pyhfst"
		import pyhfst as hfst




api_url = "https://akusanat.com/smsxml/"
download_server_url = "http://models.uralicnlp.com/nightly/"

class ModelNotFound(Exception):
    pass

class UnsupportedModel(Exception):
	pass

class HFSTRequired(Exception):
	pass

def __model_base_folders():
	a = os.path.join(os.path.dirname(__file__), "models")
	b = os.path.join(os.path.expanduser("~"), ".uralicnlp")
	return a, b

def __find_writable_folder(folders):
	for folder in folders:
		try:
			if not os.path.exists(folder):
				os.makedirs(folder)
			if os.access(folder, os.W_OK):
				return folder
		except:
			pass
	return None

def _Cg3__where_models(language, safe=False):
	return __where_models(language, safe)

def is_language_installed(language):
	path = __where_models(language, True)
	if path is None:
		return False
	return True

def _file_modification_time(filename):
    t = os.path.getmtime(filename)
    return datetime.datetime.fromtimestamp(t)

def __where_models(language, safe=False, return_latest=True):
	paths = __model_base_folders()
	latest_model = [None, 0]
	for path in paths:
		path = os.path.join(path, language)
		e = os.path.exists(path)
		if e:
			if return_latest is False:
				return path
			else:
				files = glob.glob(os.path.join(path, "*"))
				files.sort(key=os.path.getmtime)
				if len(files) == 0:
					continue
				t = _file_modification_time(files[0])
				if latest_model[0] is None or t > latest_model[1]:
					latest_model = [path, t]
	if return_latest and latest_model[0] is not None:
		return latest_model[0]
	if safe:
		return None
	raise ModelNotFound("Models for " + language + " were not in " + " or ".join(paths) + ". Use uralicApi.download(\""+ language+"\") to download models.")

def model_info(language):
	filename = os.path.join(__where_models(language), "metadata.json")
	d = mikatools.json_load(filename)
	mikatools.print_json_help(d)

def uninstall(language):
	path = __where_models(language, safe=True)
	while path is not None:
		print("removing " + path)
		shutil.rmtree(path)
		path = __where_models(language, safe=True)

def download(language, show_progress=True):
	model_types = {"analyser":"analyser-gt-desc.hfstol", "morpher-gt-desc.hfstol": "morpher-gt-desc.hfstol", "analyzer.pt": "../neural/"+language+"_analyzer_nmt-model_step_100000.pt","generator.pt": "../neural/"+language+"_generator_nmt-model_step_100000.pt", "lemmatizer.pt": "../neural/"+language+"_lemmatizer_nmt-model_step_100000.pt", "analyser-norm":"analyser-gt-norm.hfstol","analyser-dict":"analyser-dict-gt-norm.hfstol", "generator-desc":"generator-gt-desc.hfstol", "generator-norm":"generator-gt-norm.hfstol", "generator":"generator-dict-gt-norm.hfstol", "cg":"disambiguator.bin", "metadata.json":"metadata.json", "dictionary.json":"dictionary.json"}
	download_to = os.path.join(__find_writable_folder(__model_base_folders()), language)
	ssl._create_default_https_context = ssl._create_unverified_context
	if not os.path.exists(download_to):
		os.makedirs(download_to)
	for model_type, model_url in model_types.items():
		try:
			print("Downloading " + model_type + " for " + language)
			url = download_server_url + language + "/" + model_url
			save_to = os.path.join(download_to, model_type)
			mikatools.download_file(url, save_to, show_progress)
			print("Model " + model_type + " for " + language + " was downloaded")
		except:
			if model_type == "metadata.json":
				mikatools.json_dump({"info":"no metadata provided"}, save_to)
			print("Couldn't find " + model_type + " for " + language + ". It is usually safe to ignore this message")

generator_cache = {}
analyzer_cache = {}
dictionary_cache = {}

def __generator_model_name(descriptive, dictionary_forms):
	if not descriptive and dictionary_forms:
		return "generator"
	elif descriptive:
		return "generator-desc"
	else:
		return "generator-norm"

def __generate_locally(query, language, cache=True, descriptive=False, dictionary_forms=True,filename=None):
	generator = get_transducer(language,cache=cache, analyzer=False, descriptive=descriptive, dictionary_forms=dictionary_forms,filename=filename,force_no_list=False)
	if not isinstance(generator, list):
		generator = [generator]
	r = []
	[r.extend(x.lookup(query)) for x in generator]
	return r

def get_transducer(language, cache=True, analyzer=True, descriptive=True, dictionary_forms=True, convert_to_openfst=False, filename=None, force_no_list=True, segmentation=False):
	if convert_to_openfst and hfst_backend == "pyhfst":
		raise HFSTRequired("OpenFST conversion requires pip install hfst or pip install hfst-dev")
	if not analyzer:
		#generator
		if filename is None:
			filename = os.path.join(__where_models(language), __generator_model_name(descriptive, dictionary_forms))
		if cache and filename in generator_cache:
			generator = generator_cache[filename]
		else:
			generator = _load_transducer(filename, True)
			if convert_to_openfst:
				conversion_type = hfst.ImplementationType.TROPICAL_OPENFST_TYPE
				if  isinstance(generator, list):
					[x.convert(conversion_type) for x in generator]
				else:
					generator.convert(conversion_type)
			generator_cache[filename] = generator
	else:
		if filename is None:
			filename = os.path.join(__where_models(language), __analyzer_model_name(descriptive, dictionary_forms, segmentation=segmentation))
		if cache and filename in analyzer_cache:
			generator = analyzer_cache[filename]
		else:
			generator =  _load_transducer(filename, False)
			if convert_to_openfst:
				conversion_type = hfst.ImplementationType.TROPICAL_OPENFST_TYPE
				if  isinstance(generator, list):
					[x.convert(conversion_type) for x in generator]
				else:
					generator.convert(conversion_type)
			analyzer_cache[filename] = generator
	if force_no_list and isinstance(generator, list):
		generator = generator[0]
	return generator

def _load_transducer(filename, invert):
	metadata_filename =  os.path.join(os.path.dirname(filename), "metadata.json")
	try:
		metadata = mikatools.json_load(metadata_filename)
	except:
		#No crash if JSON is not found or malformed for some reason
		metadata = {}
	if "fst_type" in metadata and metadata["fst_type"] == "foma":
		return FomaFSTWrapper(filename, invert)
	elif "fst_type" in metadata and metadata["fst_type"] == "att":
		return hfst.AttReader(mikatools.open_read(filename)).read()
	elif "apertium" in metadata and metadata["apertium"] == True and hfst_backend != "pyhfst":
		input_stream = hfst.HfstInputStream(filename)
		parts = []
		while not input_stream.is_eof():
			parts.append(input_stream.read())
		return parts #input_stream.read_all()
	else:
		if "apertium" in metadata and metadata["apertium"] == True and hfst_backend == "pyhfst":
			print("Apertium transducers may require you to pip install hfst or pip install hfst-dev")
		input_stream = hfst.HfstInputStream(filename)
		return input_stream.read()



def __analyzer_model_name(descriptive, dictionary, segmentation=False):
	if segmentation:
		return "morpher-gt-desc.hfstol"
	if dictionary:
		return "analyser-dict"
	elif descriptive:
		return "analyser"
	else:
		return "analyser-norm"

def __analyze_locally(query, language, cache=True, descriptive=True, dictionary_forms=False, filename=None,segmentation=False):
	generator = get_transducer(language,cache=cache, analyzer=True, descriptive=descriptive, dictionary_forms=dictionary_forms,filename=filename, force_no_list=False,segmentation=segmentation)
	if not isinstance(generator, list):
		generator = [generator]
	r = []
	[r.extend(x.lookup(query)) for x in generator]
	return r

def __encode_query(query):
	if not new_python and type(query) is unicode:
		query = query.encode('utf-8')
	return query

def __regex_escape(word):
	escape = ["#", "/","+","~","\\","&","-","$", "*", "|", "?", "{","}","\"",":",";","!", "."]
	for e in escape:
		word = word.replace(e, "%" +e)
	return word

def _get_flag_string(abcs, filter_out):
	f = []
	flags = []
	for abc in abcs:
		for fi in filter_out:
			if abc.startswith(fi):
				f.append(__regex_escape(abc))
				break
		if "@" in abc and "@_"  not in abc:
			flags.append( "\"" + abc + "\"")
	flag_string = ""
	flag_end = ""
	start_flag_end = ""
	flag_string_start = ""
	if len(flags) == 0:
		flags = ['"@DUMMY@"']
	if len(flags) > 0:
		flag_string_start =  " [ "+ " | ".join(flags)
		flag_string =  flag_string_start +" | "
		flag_string_start = "" +flag_string_start
		flag_end = "]"
		start_flag_end = "]* "
	return flag_string_start, start_flag_end, flag_string, flag_end, f

def get_all_forms(word, pos, language, descriptive=True, limit_forms=-1, filter_out=["#", "+Der", "+Cmp","+Err"]):
	if hfst_backend == "pyhfst":
		HFSTRequired("get_all_forms requires pip install hfst or pip install hfst-dev")
	analyzer = get_transducer(language, descriptive=descriptive, analyzer=True, convert_to_openfst=True, cache=True, dictionary_forms=False, force_no_list=False)
	if isinstance(analyzer, list):
		#apertium
		analyzers = analyzer
		out = []
		for analyzer in analyzers:
			abcs = analyzer.get_alphabet()
			flag_string_start, start_flag_end, flag_string, flag_end, f = _get_flag_string(abcs, filter_out)
			if len(f) > 0:
				f_string = " - [ "+ " | ".join(f) +" ]"
			else:
				f_string = ""
			reg_text = flag_string_start + start_flag_end + "{"+word+"} %<"+pos+ "%>" + flag_string + " [ ? "+f_string+"]"+flag_end+"*"
			reg = hfst.regex(reg_text)
			analyzer2 = analyzer.copy()
			analyzer2.compose(reg)
			output = analyzer2.extract_paths(max_cycles=1, max_number=limit_forms,output='text').replace("@_EPSILON_SYMBOL_@","").split("\n")
			output = [_o.split('\t') for _o in output if _o]
			output = [(":".join(_o[:-1]), float(_o[-1]), ) for _o in output]
			out.extend(output)
		return out

	else:
		abcs = analyzer.get_alphabet()
		flag_string_start, start_flag_end, flag_string, flag_end, f = _get_flag_string(abcs, filter_out)
		reg_text = flag_string_start + start_flag_end + "{"+word+"} %+"+pos+ flag_string + " [ ? -  [ "+ " | ".join(f) +" ]]"+flag_end+"*"
		reg = hfst.regex(reg_text)
		analyzer2 = analyzer.copy()
		analyzer2.compose(reg)
		output = analyzer2.extract_paths(max_cycles=1, max_number=limit_forms,output='text').replace("@_EPSILON_SYMBOL_@","").split("\n")
		output = [_o.split('\t') for _o in output if _o]
		output = [(":".join(_o[:-1]), float(_o[-1]), ) for _o in output]
		return output

def generate(query, language, force_local=True, descriptive=False, dictionary_forms=False, remove_symbols=True, filename=None, neural_fallback=False, n_best=1):
	if force_local or __where_models(language, safe=True):
		r = __generate_locally(__encode_query(query), language, descriptive=descriptive, dictionary_forms=dictionary_forms,filename=filename)
	else:
		r = __api_generate(query, language, descriptive=descriptive, dictionary_forms=dictionary_forms)
	if remove_symbols:
		r = _remove_analysis_symbols(r)
	if neural_fallback and len(r) == 0:
		nfst = NeuralFST(__where_models(language, safe=True))
		return nfst.generate(query, n_best=n_best)
	return r

def __remove_symbols(string):
	return re.sub('@[^@]*@', '', string)

def analyze(query, language, force_local=True, descriptive=True, remove_symbols=True,language_flags=False, dictionary_forms=False,filename=None,neural_fallback=False, n_best=1, segmentation=False):
	if not isinstance(language, str) and isinstance(language, Iterable):
		#Treat as a list
		r = []
		for l in language:
			r.extend(analyze(query,l, force_local=force_local, descriptive=descriptive, remove_symbols=remove_symbols,language_flags=language_flags, dictionary_forms=dictionary_forms,filename=filename))
		language_flags = False #don't add them again

	elif force_local or __where_models(language, safe=True):
		r = __analyze_locally(__encode_query(query), language,descriptive=descriptive,dictionary_forms=dictionary_forms,filename=filename,segmentation=segmentation)
	else:
		r = __api_analyze(query, language,descriptive=descriptive)
	if remove_symbols:
		r = _remove_analysis_symbols(r)
	if neural_fallback and len(r) == 0:
		nfst = NeuralFST(__where_models(language, safe=True))
		r = nfst.analyze(query, n_best=n_best)

	if language_flags:
		return _add_language_flag(r, language)
	else:
		return r

def _add_language_flag(analysis, language):
	r = []
	for a in analysis:
		r.append((a[0] + "+" +language, a[1]))
	return r

def _remove_analysis_symbols(r):
	r = list(r)
	for i in range(len(r)):
		item = r[i]
		r[i] = (__remove_symbols(item[0]),item[1])
	return r

def lemmatize(word, language, force_local=True, descriptive=True, word_boundaries=False, dictionary_forms=False, filename=None, neural_fallback = False, n_best=1):
    analysis = analyze(word, language, force_local, descriptive=descriptive, dictionary_forms=dictionary_forms, filename=filename,neural_fallback=neural_fallback, n_best=n_best)
    lemmas = []
    if word_boundaries:
    	bound = "|"
    else:
    	bound = ""
    for tupla in analysis:
        an = tupla[0]
        if language == "swe":
            lemma = re.sub("[\<].*?[\>]", bound, an).strip(bound)
            lemmas.append(lemma)
        elif language == "ara":
        	lemmas.append(filter_arabic(an,combine_by=bound))
        elif language == "fin_hist":
        	lemma = bound.join(re.findall("(?<=WORD_ID=)[^\]]*", an))
        	lemmas.append(lemma)
        elif "<" in an and ">" in an:
        	#apertium
        	parts = an.split("+")
        	lemma = bound.join([x.split("<")[0] for x in parts])
        	lemmas.append(lemma)
        else:
            if not "+Cmp#" in an and "#" in an:
                an = an.replace("#", "+Cmp#")
            res = an.split("+Cmp#")
            lemma = [x.split("+")[0] for x in res]
            if language == "eng":
            	lemma = [re.sub("[\[].*?[\]]", "", x) for x in lemma]
            lemmas.append(bound.join(lemma))

    lemmas = list(set(lemmas))
    return lemmas

def supported_languages():
	d = requests.get(download_server_url  + "supported_languages.json")
	return d.json()

def __api_analyze(word, language,descriptive=True):
	if not descriptive:
		raise UnsupportedModel("Server only supports descriptive analysis. Please download the models and analyze locally")
	return __send_request("analyze/", {"word": word, "language": language})["analysis"]

def __api_generate(query, language, descriptive=False, dictionary_forms=True):
	if descriptive or not dictionary_forms:
		raise UnsupportedModel("Server only supports normative dictionary forms. Please download the models and generate locally")
	return __send_request("generate/", {"query": query, "language": language})["analysis"]

def dictionary_search(word, language, force_local=True, backend=TinyDictionary):
	if force_local:
		d = _get_dictionary(language,backend=backend)
		lemmas = list(set(lemmatize(word, language)) - set([word]))
		return d.find(word, lemmas)
	else:
		return _api_dictionary_search(word, language)

def dictionary_lemmas(language, backend=TinyDictionary, group_by_pos=False):
	d = _get_dictionary(language,backend=backend)
	return d.lemmas(group_by_pos=group_by_pos)

def import_dictionary_to_db(language, backend=MongoDictionary):
	d = _get_dictionary(language, backend=backend)
	d.import_data()

def _api_dictionary_search(word, language):
	return __send_request("search/", {"word": word, "language": language})

def __send_request(url, data):
	r = requests.get(api_url + url, params=data)
	return r.json()

def _get_dictionary(language, backend=TinyDictionary):
	cache_key = backend.__name__ + "_" + language
	if cache_key in dictionary_cache:
		return dictionary_cache[cache_key]
	path = os.path.join(__where_models(language), "dictionary.json")
	dictionary = backend(path, language)
	dictionary_cache[cache_key] = dictionary
	return dictionary

def segment(query, language):
	return [x[0].replace("#",">").split(">") for x in analyze(query, language, segmentation=True)]


def get_translation(lemma, lang, trans_lang, backend=TinyDictionary):
	res = dictionary_search(lemma, lang, backend=backend)
	translations = []
	for e in res["exact_match"]:
		mgs = e["mg"]
		if not isinstance(mgs, list):
			mgs = [mgs]
		for mg in mgs:
			tgs = mg["tg"]
			if not isinstance(tgs, list):
				tgs = [tgs]
			for tg in tgs:
				for tr in tgs:
					if tr["@xml:lang"] == trans_lang:
						ts = tr["t"]
						if not isinstance(ts, list):
							ts = [ts]
						for t in ts:
							translations.append(t["#text"])
		translations = list(set(translations))
	return translations




#encoding: utf-8
from __future__ import unicode_literals
import re, unicodedata
import mikatools

isos = mikatools.json_load(mikatools.script_path("lang_codes.json"))

pattern = re.compile(r'(\w[\u02F3\u0300\u2013\u032E\u208D\u203F\u0311\u0323\u035E\u031C\u02FC\u030C\u02F9\u0328\u032D:\u02F4\u032F\u0330\u035C\u0302\u0327\u03572\u0308\u0351\u0304\u02F2\u0352\u0355\u00B7\u032C\u030B\u2019\u0339\u00B4\u0301\u02F1\u0303\u0306\u030A7\u0325\u0307\u0354`\u02F0]+|\w|\W)', re.UNICODE | re.IGNORECASE)

def char_split(word):
	word = unicodedata.normalize('NFKC',word)
	_result = pattern.findall(word)
	return list(_result)

def filter_arabic(text, keep_vowels=True, combine_by=""):
	if keep_vowels:
		return combine_by.join(re.findall(r"[ء-ي'ًٌٍَُِّْـ']+", text))
	else:
		return combine_by.join(re.findall(r"[ء-ي]+", text))
		
def iso_to_name(iso):
    return isos[iso]

arabic_diacritics = re.compile("""
                             ّ    | # Tashdid
                             َ    | # Fatha
                             ً    | # Tanwin Fath
                             ُ    | # Damma
                             ٌ    | # Tanwin Damm
                             ِ    | # Kasra
                             ٍ    | # Tanwin Kasr
                             ْ    | # Sukun
                             ـ     # Tatwil/Kashida
                         """, re.VERBOSE)


def remove_arabic_diacritics(text):
    text = re.sub(arabic_diacritics, '', text)
    return text


try:
	from . import foma
except:
	foma = None
import warnings

arabic_clitics = {"art":{"ال":["ال"]}, "prep":{"ب":["بِ"], "ل":["لِ","لَ"], "ت": ["تَ"], "ك":["كَ"], "س":["سَ"]}, "conj":{"ف":["فَ"], "و":["وَ"]}}

arabic_vowels = "َ"

class FomaFSTWrapper(object):
	"""docstring for FomaFSTWrapper"""
	def __init__(self, filename, invert):
		if foma is None:
			raise Exception("Foma (libfoma) is not installed or the version is too old, please refer to https://fomafst.github.io/")
		self.foma = foma.FST.load(filename)
		if invert:
			self.lookup = self.generate
		else:
			self.lookup = self.analyze
		if "\"" in filename:
			s = "\""
		else:
			s = "/"
		file_split = filename.split(s)
		if len(file_split) > 2 and file_split[-2] == "ara":
			self.arab = True
		else:
			self.arab = False
			

	def generate(self, query):
		if self.arab:
			return self._arab_generate(query)
		return self._dummy_weights(self.foma.apply_down(query))

	def analyze(self, query):
		if self.arab:
			return self._arab_analyze(query)
		return self._dummy_weights(self.foma.apply_up(query))

	def _arabic_trim_vowel(self, word):
		for x in arabic_vowels:
			if word.startswith(x):
				word = word[1:]
		return word

	def _add_to_arab_results(self, results, voweled_prefixes):
		res = []
		if voweled_prefixes is None:
			return results
		for vp in voweled_prefixes:
			res.extend([vp + "#" + x[0] for x in results])
		return res
	def _arab_analyze(self, query, recurse=True):
		res = []
		if recurse:
			word = query
			set_conj = None
			set_prep = None
			for conj, voweled in arabic_clitics["conj"].items():
				if word.startswith(conj) and word is not conj:
					set_conj = voweled
					word = self._arabic_trim_vowel(word[len(conj):])
					res.extend( self._add_to_arab_results( self._arab_analyze(word, recurse=False), set_conj ) )
					break
			for prep, voweled in arabic_clitics["prep"].items():
				if word.startswith(prep) and word is not prep:
					set_prep = voweled
					word = self._arabic_trim_vowel(word[len(prep):])
					res.extend( self._add_to_arab_results(self._add_to_arab_results( self._arab_analyze(word, recurse=False), set_prep ), set_conj) )
					break
			for art, art_voweled in arabic_clitics["art"].items():
				if word.startswith(art) and art is not word:
					word = self._arabic_trim_vowel(word[len(art):])
					res.extend( self._add_to_arab_results(self._add_to_arab_results(self._add_to_arab_results( self._arab_analyze(word, recurse=False), art_voweled ), set_prep), set_conj))
					break
		res.extend(self.foma.apply_up(query))
		res = list(set(res))
		return self._dummy_weights(res)

	def _arab_generate(self, query, recurse=True):
		return self._dummy_weights(self.foma.apply_down(query))

	def _dummy_weights(self, res):
		res = list(res)
		return [(x, 0.0) for x in res]

	def convert(self, *args, **kwargs):
		warnings.warn("Conversion not supported by Foma based transducers")



try:
	from natas.normalize import call_onmt
except:
	call_onmt = None
import os
class NatasNotInstalled(Exception):
	pass

class NeuralFST(object):
	"""docstring for NeuralFST"""
	def __init__(self, model_path):
		if call_onmt is None:
			raise NatasNotInstalled("Natas is needed for neural models, run:\n\npip install natas")
		
		self.model_path = model_path
	
	def analyze(self, word, n_best=1):
		if len(word) == 0:
			return []
		model_a = os.path.join(self.model_path, "analyzer.pt")
		model_l = os.path.join(self.model_path, "lemmatizer.pt")
		tags = call_onmt([" ".join(word)] ,model_a,n_best=n_best,return_scores=True)[0]
		lemma = call_onmt([" ".join(word)], model_l,n_best=n_best,return_scores=True)[0]
		lem_tags = zip(tags, lemma)
		return [(l[0].replace(" ","") + "+" + t[0].replace(" ","+"), l[1] + t[1]) for t,l in lem_tags ]

	def generate(self, word, n_best=1):
		if len(word) ==0:
			return []
		model_g = os.path.join(self.model_path, "generator.pt")
		parts = word.split("+")
		parts[0] = " ".join(parts[0])
		forms =  call_onmt([" ".join(parts)] ,model_g, n_best=n_best,return_scores=True)[0]
		return [(form.replace(" ", ""), score) for form, score in forms]




import re
import mikatools
from . import uralicApi
from . import string_processing

preps = list(set("كبولف"))
al = "ال"

sentence_end = set("!?。……‥！？。⋯…؟჻!…")
word_end_puct = set(",;:”’'\"»」)]}،؛》』〕｠〉》】〗〙〛–—")
word_start_punct = set("'\"¡¿「«“”‘({[《『〔｟〈《【〖〘〚–—”")
numbers = set("0123456789١٢٣٤٥٦٧٨٩٠")
_custom_punctuation = set('!"#$%&\'()*+,-.:;<=>?@[]^_`{|}~')

abrvs = mikatools.json_load(mikatools.script_path("abrvs.json"))
abrv_regex = re.compile(r"(^|\s)(" + "|".join([re.escape(a) for a in abrvs]) + ")$")

def _ends_in_abrv(text):
	return abrv_regex.search(text.lower()) is not None

def _remove_preps(word):
	for p in preps:
		if word.startswith(p):
			word = word[1:]
			return word, p
	return word, None

def tokenize_arabic(text):
	toksut = sentences(text)
	#print(toksut)
	output = []
	for sentence in toksut:
		o_sentence = []
		for w in words(sentence):
			#print(o_sentence)
			#print(w)
			word = w
			lemmas = uralicApi.lemmatize(word, "ara",word_boundaries=True)
			preppu = None
			alli = None
			high_count = 0
			long_len = 0
			stop =False
			for l in lemmas:
				if "|" in l:
					c = l.count("|")
					if c > high_count:
						high_count = c
						long_len = len(l)
						stop = l
					elif high_count == c:
						if len(l) > long_len:
							stop = l
							long_len = len(l)
			if high_count > 0:
				o_sentence.extend([x for x in stop.split("|") if len(string_processing.remove_arabic_diacritics(x)) >0])
				continue

			if len(lemmas) == 0:
				word, preppu = _remove_preps(word)
				lemmas = uralicApi.lemmatize(word, "ara")
			if len(lemmas) == 0 and word.startswith(al):
				word = word[2:]
				alli = al
				lemmas = uralicApi.lemmatize(word, "ara")
			if len(lemmas) == 0:
				o_sentence.append(w)
			else:
				parttusan = [preppu, alli, lemmas[0]]
				#print(parttusan)
				o_sentence.extend([x for x in parttusan if x is not None])
		output.append(o_sentence)
	return output

def sentences(text):
	parts = []
	current_s = ""
	previous_break = False
	for i, c in enumerate(text):
		if c in sentence_end:
			#End of a sentence, not a dot
			if len(current_s) > 0:
				#There is a current sentence, apped it and clear it
				parts.append(current_s + c)
				current_s = ""
			elif len(parts) > 0:
				#No current sentence, add to a previous sentence
				parts[len(parts)-1] += c
			else:
				#Make it a current sentence
				current_s = c
		elif c == ".":
			#A dot
			if len(current_s) == 0:
				#no current sentence
				if len(parts) > 0:
					#append
					parts[len(parts)-1] += c
				else:
					current_s = c
			elif len(current_s) > 0 and current_s[-1] in numbers:
				#previous is a number
				current_s += c
			elif _ends_in_abrv(current_s):
				#abreviation
				current_s += c
			elif len(text) > i+1 and len(text[i+1].strip()) != 0:
				#dot is not followed by a space
				current_s += c
			else:
				#dot ending a sentence
				parts.append(current_s + c)
				current_s = ""
		elif c == "\n":
			#line break
			if previous_break and len(current_s) > 0:
				parts.append(current_s)
				current_s = ""
			if not previous_break and len(current_s) > 0:
				current_s += c
			previous_break = True
			continue
		elif c == "\r":
			#Windows line break
			continue
		else:
			#Any other character
			current_s += c
		previous_break = False
	if len(current_s) > 0:
		parts.append(current_s)
	space_regex = re.compile(r"\s+")
	parts = [space_regex.sub(" ", part).strip() for part in parts]
	parts = [part for part in parts if len(part) > 0]
	return parts


def words(text):
	multidot = re.compile(r"(\.{2,})$")
	space_regex = re.compile(r"\s+")
	for sentence_end_p in sentence_end:
		text = text.replace(sentence_end_p, " " + sentence_end_p)
	text = space_regex.sub(" ", text).strip()
	whitespace_tokens = text.split(" ")
	tokens = []
	for t in whitespace_tokens:
		first_tok = []
		last_tok = []
		cont_first = True
		while cont_first:
			cont_first = False
			if len(t) > 0 and t[0] in word_start_punct:
				cont_first = True
				first_tok.append(t[0])
				t = t[1:]
		cont_last = True
		while cont_last:
			cont_last = False
			if len(t) > 0 and t[-1] in word_end_puct:
				cont_last = True
				last_tok.insert(0,t[-1])
				t = t[:-1]
			elif len(t)>1 and t[-1] == "." and t[-2] in word_end_puct:
				cont_last = True
				last_tok.insert(0,t[-1])
				last_tok.insert(0,t[-2])
				t = t[:-2]
		dots = multidot.search(t)
		if dots is not None:
			#Make .. or ... or whaterver its own token
			dots = dots.groups()[0]
			last_tok.insert(0, dots)
			t = t[:-len(dots)]
		elif len(t) >0 and t[-1] == ".":
			if not _ends_in_abrv(t[:-1]):
				t = t[:-1]
				last_tok.insert(0, ".")

		if ("/" in t or "\\" in t) and not any((x in _custom_punctuation for x in t)):
			#not a link
			t = t.replace("/", " /").replace("\\", " \\")
			t = t.split(" ")
		else:
			t = [t]
		t = [x for x in t if len(x) > 0]

		first_tok.extend(t)
		first_tok.extend(last_tok)
		tokens.extend(first_tok)

	tokens = [tok for tok in tokens if len(tok) > 0]
	return tokens

	

def tokenize(text):
	sents = sentences(text)
	return [words(s) for s in sents]


import requests
from collections import defaultdict
from uralicNLP.dependency import LanguageNotSupported


class Traslator(object):
	"""docstring for Traslator"""
	def __init__(self):
		super(Traslator, self).__init__()
		self.languages = None

	def translate(self, text, source, target):
		self._get_languages()
		if source not in self.languages:
			raise LanguageNotSupported(source + " is not a supported language")
		if target not in self.languages[source]:
			raise LanguageNotSupported(source + "-" + target + " is not a supported pair")
		return self._translate(text, source, target)


	def get_languages(self, force_refresh=False):
		if self.languages == None or force_refresh:
			self._get_languages()
		return self.languages

class TartuTranslator(Traslator):

	def __init__(self, domain="general"):
		super(TartuTranslator, self).__init__()
		self.url = "https://api.tartunlp.ai/translation/v2/"
		self.domain = domain

	def _get_languages(self):
		r = requests.get(self.url)
		data = r.json()["domains"]
		res = {}
		for d in data:
			d_res = {}
			res[d["name"]] = d_res
			for lang_pair in d["languages"]:
				lang1, lang2 = lang_pair.split("-")
				if lang1 not in d_res:
					d_res[lang1] = []
				d_res[lang1].append(lang2)

			res[d["name"]] = d_res
		self.languages = res

	def translate(self, text, source, target):
		r = requests.post(self.url, json={"text":text, "src": source, "tgt": target, "domain": self.domain, "application": "uralicNLP"})
		if r.status_code != 200:
			raise LanguageNotSupported(r.text)
		data = r.json()
		return data["result"]




class ApertiumTranslator(Traslator):
	"""docstring for ApertiumTranslator"""
	def __init__(self, url):
		super(ApertiumTranslator, self).__init__()
		self.url = url

	def _get_languages(self):
		r = requests.get(self.url + "list?q=pairs")
		data = r.json()
		languages = defaultdict(list)
		for item in data["responseData"]:
			languages[item["sourceLanguage"]].append(item["targetLanguage"])
		self.languages = languages

	def _translate(self, text, source, target):
		r = requests.post(self.url + "translate", {"langpair": source + "|" + target,"q": text})
		data = r.json()
		return data["responseData"]["translatedText"]


		
class YandexTranslator(Traslator):
	"""docstring for YandexTranslator"""
	def __init__(self, api_key):
		super(YandexTranslator, self).__init__()
		self.api_key = api_key

	def _get_languages(self):
		r = requests.get("https://translate.yandex.net/api/v1.5/tr.json/getLangs", {"key": self.api_key, "ui":"en"})
		data = r.json()
		langs = defaultdict(list)
		for item in data["dirs"]:
			s, t = item.split("-")
			langs[s].append(t)
		self.languages = langs

	def _translate(self, text, source, target):
		r = requests.post("https://translate.yandex.net/api/v1.5/tr.json/translate", {"key": self.api_key, "text":text, "lang": source + "-" + target})
		data = r.json()
		return data["text"][0]

	

def ApertiumBetaTranslator():
	return ApertiumTranslator("http://beta.apertium.org/apy/")

def ApertiumStableTranslator():
	return ApertiumTranslator("https://www.apertium.org/apy/")

def ApertiumGiellateknoTranslator():
	return ApertiumTranslator("https://gtweb.uit.no/apy/")

ApertiumJorgalTranslator = ApertiumGiellateknoTranslator

def ApertiumLocalhostTranslator():
	return ApertiumTranslator("http://localhost:2737/")





# -*- coding: utf-8 -*-

#   Foma: a finite-state toolkit and library.                                 #
#   Copyright © 2008-2015 Mans Hulden                                         #

#   This file is part of foma.                                                #

#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #

#      http://www.apache.org/licenses/LICENSE-2.0                             #

#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #

from past.builtins import xrange
from sys import maxsize
from ctypes import *
from ctypes.util import find_library

from . import six_1_11_0 as six
import warnings

fomalibpath = find_library('foma')
foma = cdll.LoadLibrary(fomalibpath)


class FSTstruct(Structure):
    _fields_ = [
        ("name", c_char * 40),
        ("arity", c_int),
        ("arccount", c_int),
        ("statecount", c_int),
        ("linecount", c_int),
        ("finalcount", c_int),
        ("pathcount", c_longlong),
        ("is_deterministic", c_int),
        ("is_pruned", c_int),
        ("is_minimized", c_int),
        ("is_epsilon_free", c_int),
        ("is_loop_free", c_int),
        ("is_completed", c_int),
        ("arcs_sorted_in", c_int),
        ("arcs_sorted_out", c_int),
        ("fsm_state", c_void_p),
        ("sigma", c_void_p),
        ("medlookup", c_void_p)
    ]


foma_fsm_parse_regex = foma.fsm_parse_regex
foma_fsm_parse_regex.restype = POINTER(FSTstruct)
foma_apply_init = foma.apply_init
foma_apply_init.restype = c_void_p
foma_apply_clear = foma.apply_clear
foma_apply_words = foma.apply_words
foma_apply_words.restype = c_char_p
foma_apply_lower_words = foma.apply_lower_words
foma_apply_lower_words.restype = c_char_p
foma_apply_upper_words = foma.apply_upper_words
foma_apply_upper_words.restype = c_char_p
foma_apply_down = foma.apply_down
foma_apply_down.restype = c_char_p
foma_apply_up = foma.apply_up
foma_apply_up.restype = c_char_p
foma_apply_set_space_symbol = foma.apply_set_space_symbol
foma_fsm_count = foma.fsm_count
foma_fsm_topsort = foma.fsm_topsort
foma_fsm_topsort.restype = POINTER(FSTstruct)
foma_fsm_minimize = foma.fsm_minimize
foma_fsm_minimize.restype = POINTER(FSTstruct)
foma_fsm_union = foma.fsm_union
foma_fsm_union.restype = POINTER(FSTstruct)
foma_fsm_intersect = foma.fsm_intersect
foma_fsm_intersect.restype = POINTER(FSTstruct)
foma_fsm_minus = foma.fsm_minus
foma_fsm_minus.restype = POINTER(FSTstruct)
foma_fsm_compose = foma.fsm_compose
foma_fsm_compose.restype = POINTER(FSTstruct)
foma_fsm_concat = foma.fsm_concat
foma_fsm_concat.restype = POINTER(FSTstruct)
foma_fsm_copy = foma.fsm_copy
foma_fsm_copy.restype = POINTER(FSTstruct)
foma_fsm_complement = foma.fsm_complement
foma_fsm_complement.restype = POINTER(FSTstruct)
foma_fsm_lower = foma.fsm_lower
foma_fsm_lower.restype = POINTER(FSTstruct)
foma_fsm_upper = foma.fsm_upper
foma_fsm_upper.restype = POINTER(FSTstruct)
foma_fsm_minimize = foma.fsm_minimize
foma_fsm_minimize.restype = POINTER(FSTstruct)
foma_fsm_destroy = foma.fsm_destroy
foma_fsm_equivalent = foma.fsm_equivalent
foma_fsm_equivalent.restype = c_int
foma_fsm_isempty = foma.fsm_isempty
foma_fsm_isempty.restype = c_int
foma_fsm_flatten = foma.fsm_flatten
foma_fsm_flatten.restype = POINTER(FSTstruct)
foma_apply_set_space_symbol = foma.apply_set_space_symbol
foma_fsm_read_binary_file = foma.fsm_read_binary_file
foma_fsm_read_binary_file.restype = POINTER(FSTstruct)


"""Define functions."""
foma_add_defined = foma.add_defined
foma_add_defined.restype = c_int
try:
    #These seem to cause a crash on some systems...
    foma_add_defined_function = foma.add_defined_function
    foma_add_defined_function.restype = c_int
except:
    pass
defined_networks_init = foma.defined_networks_init
defined_networks_init.restype = c_void_p
defined_functions_init = foma.defined_functions_init
defined_functions_init.restype = c_void_p

"""Trie functions."""
fsm_trie_init = foma.fsm_trie_init
fsm_trie_init.restype = c_void_p
fsm_trie_add_word = foma.fsm_trie_add_word
fsm_trie_done = foma.fsm_trie_done
fsm_trie_done.restype = POINTER(FSTstruct)


class FSTnetworkdefinitions(object):
    def __init__(self):
        self.defhandle = defined_networks_init(None)


class FSTfunctiondefinitions(object):
    def __init__(self):
        self.deffhandle = defined_functions_init(None)


class FST(object):
    networkdefinitions = FSTnetworkdefinitions()
    functiondefinitions = FSTfunctiondefinitions()

    @classmethod
    def define(cls, definition, name):
        """Defines an FSM constant; can be supplied regex or existing FSM."""
        name = cls.encode(name)
        if isinstance(definition, FST):
            retval = foma.add_defined(c_void_p(cls.networkdefinitions.defhandle), foma_fsm_copy(definition.fsthandle), c_char_p(name))
        elif isinstance(definition, six.string_types):
            regex = cls.encode(definition)
            retval = foma.add_defined(c_void_p(cls.networkdefinitions.defhandle), foma_fsm_parse_regex(c_char_p(regex), c_void_p(cls.networkdefinitions.defhandle), c_void_p(cls.functiondefinitions.deffhandle)), c_char_p(name))
        else:
            raise ValueError("Expected str, unicode, or FSM")

    @classmethod
    def definef(cls, prototype, definition):
        """Defines an FSM function."""
        # Prototype is a 2-tuple (name, (arg1name, ..., argname))
        # Definition is regex using prototype variables
        name = cls.encode(prototype[0] + '(')
        if isinstance(definition, six.string_types):
            numargs = len(prototype[1])
            for i in xrange(numargs):
                definition = definition.replace(prototype[1][i], "@ARGUMENT0%i@" % (i+1))
            regex = cls.encode(definition + ';')
            retval = foma.add_defined_function(c_void_p(cls.functiondefinitions.deffhandle), c_char_p(name), c_char_p(regex), c_int(numargs))
        else:
            raise ValueError("Expected regex as definition")
         
    @classmethod
    def wordlist(cls, wordlist, minimize = True):
        """Create FSM directly from wordlist.
           Returns a trie-shaped deterministic automaton if not minimized."""
        th = fsm_trie_init()
        for w in wordlist:
            thisword = cls.encode(w)
            fsm_trie_add_word(c_void_p(th), c_char_p(thisword))
        fsm = cls()
        fsm.fsthandle = fsm_trie_done(c_void_p(th))
        if minimize:
            fsm.fsthandle = foma_fsm_minimize(fsm.fsthandle)
        return fsm

    @classmethod
    def load(cls, filename):
        """Load binary FSM from file."""
        fsm = cls()
        fsm.fsthandle = foma_fsm_read_binary_file(c_char_p(FST.encode(filename)))
        if not fsm.fsthandle:
            raise ValueError("File error.")
        return fsm

    @staticmethod
    def encode(string):
        # type: (Any) -> six.binary_type
        """Makes sure str and unicode are converted."""
        if isinstance(string, six.text_type):
            return string.encode('utf8')
        elif isinstance(string, six.binary_type):
            return string
        else:
            return FST.encode(str(string))

    @staticmethod
    def decode(text):
        if text is None:
            return None
        elif isinstance(text, six.binary_type):
            # Assume output is UTF-8 encoded:
            return text.decode('UTF-8')
        else:
            assert isinstance(text, six.text_type)
            return text

    def __init__(self, regex = False):
        if regex:
            self.regex = self.encode(regex)
            self.fsthandle = foma_fsm_parse_regex(c_char_p(self.regex), c_void_p(self.networkdefinitions.defhandle), c_void_p(self.functiondefinitions.deffhandle))
            if not self.fsthandle:
                raise ValueError("Syntax error in regex")
        else:
            self.fsthandle = None
        self.getitemapplyer = None

    def __getitem__(self, key):
        if not self.fsthandle:
            raise KeyError('FST not defined')
        if not self.getitemapplyer:
            self.getitemapplyer = foma_apply_init(self.fsthandle)
        result = []
        output = foma_apply_down(c_void_p(self.getitemapplyer), c_char_p(self.encode(key)))
        while True:
            if output == None:
                return result
            else:
                result.append(output)
                output = foma_apply_down(c_void_p(self.getitemapplyer), None)
            
    def __del__(self):
        if self.fsthandle:
            foma_fsm_destroy(self.fsthandle)

    def __str__(self):
        if not self.fsthandle:
            raise ValueError('FSM not defined')
        foma_fsm_count(self.fsthandle)
        s  = 'Name: %s\n' % self.fsthandle.contents.name
        s += 'States: %i\n' % self.fsthandle.contents.statecount
        s += 'Transitions: %i\n' % self.fsthandle.contents.arccount
        s += 'Final states: %i\n' % self.fsthandle.contents.finalcount
        s += 'Deterministic: %i\n' % self.fsthandle.contents.is_deterministic
        s += 'Minimized: %i\n' % self.fsthandle.contents.is_minimized
        s += 'Arity: %i\n' % self.fsthandle.contents.arity
        return s

    def __len__(self):
        if self.fsthandle:
            if self.fsthandle.contents.pathcount == -3: # UNKNOWN
                self.fsthandle = foma_fsm_topsort(self.fsthandle)
            if self.fsthandle.contents.pathcount == -1: # CYCLIC
                raise ValueError("FSM is cyclic")
            if self.fsthandle.contents.pathcount == -2: # OVERFLOW
                return maxsize
            return self.fsthandle.contents.pathcount
        else:
            raise ValueError("FSM not defined")
        
    def __add__(self, other):
        return self.concat(other)

    def __sub__(self, other):
        return self.minus(other)
    
    def __le__(self, other):
        if self.fsthandle and other.fsthandle:
            return bool(c_int(foma_fsm_isempty(foma_fsm_minimize(foma_fsm_minus(foma_fsm_copy(self.fsthandle),foma_fsm_copy(other.fsthandle))))))
        else:
            raise ValueError('Undefined FST')

    def __lt__(self, other):
        if self.fsthandle and other.fsthandle:
            return (not self.__eq__(other)) and bool(c_int(foma_fsm_isempty(foma_fsm_minimize(foma_fsm_minus(foma_fsm_copy(self.fsthandle),foma_fsm_copy(other.fsthandle))))))
        else:
            raise ValueError('Undefined FST')        
        
    def __or__(self, other):
        return self.union(other)

    def __and__(self, other):
        return self.intersect(other)

    def __eq__(self, other):
        if self.fsthandle and other.fsthandle:
            return bool(c_int(foma_fsm_equivalent(foma_fsm_copy(self.fsthandle), foma_fsm_copy(other.fsthandle))))
        else:
            raise ValueError('Undefined FST')
    
    def __ne__(self, other):
        return not(self.__eq__(other))

    def __contains__(self, word):
        af = self.apply_down(word)
        try:
            i = af.next()
            return True
        except StopIteration:
            return False
                
    def __call__(self, other):
        if isinstance(other, six.string_types):
            return FST("{" + other + "}").compose(self)
        else:
            return other.compose(self)
    
    def __invert__(self):
        new = FST()
        new.fsthandle = self._fomacallunary(foma_fsm_complement)
        return new

    def __iter__(self):
        return self._apply(foma_apply_upper_words, word = None, tokenize = False)
    
    def _apply(self, applyf, word = None, tokenize = False):
        if not self.fsthandle:
            raise ValueError('FST not defined')
        applyerhandle = foma_apply_init(self.fsthandle)
        if tokenize:
            toksym = '\x07'
            foma_apply_set_space_symbol(c_void_p(applyerhandle), c_char_p(toksym))
        if word:
            output = applyf(c_void_p(applyerhandle), c_char_p(self.encode(word)))
        else:
            output = applyf(c_void_p(applyerhandle))
        while True:
            if output == None:
                foma_apply_clear(c_void_p(applyerhandle))
                return
            else:
                if tokenize:
                    yield output[:-1].split('\x07')
                else:
                    yield self.decode(output)
            if word:
                output = applyf(c_void_p(applyerhandle), None)
            else:
                output = applyf(c_void_p(applyerhandle))
                    
    def words(self, tokenize = False):
        return self._apply(foma_apply_words, word = None, tokenize = tokenize)

    def lowerwords(self, tokenize = False):
        return self._apply(foma_apply_lower_words, word = None, tokenize = tokenize)
                    
    def upperwords(self, tokenize = False):
        return self._apply(foma_apply_upper_words, word = None, tokenize = tokenize)
        
    def apply_down(self, word, tokenize = False):
        return self._apply(foma_apply_down, word = word, tokenize = tokenize)

    def apply_up(self, word, tokenize = False):
        if self.fsthandle:
            return self._apply(foma_apply_up, word = word, tokenize = tokenize)
        else:
            raise ValueError('Undefined FST')

    def _fomacallunary(self, func, minimize = True):
        if self.fsthandle:
            handle = func(foma_fsm_copy(self.fsthandle))
            if minimize:
                handle = foma_fsm_minimize(handle)
            return handle
        else:
            raise ValueError('Undefined FST')
        
    def _fomacallbinary(self, other, func, minimize = True):
        if self.fsthandle and other.fsthandle:
            handle = func(foma_fsm_copy(self.fsthandle), foma_fsm_copy(other.fsthandle))
            if minimize:
                handle = foma_fsm_minimize(handle)
            return handle
        else:
            raise ValueError('Undefined FST')
        
    def union(self, other, minimize = True):
        new = FST()
        new.fsthandle = self._fomacallbinary(other, foma_fsm_union, minimize)
        return new

    def intersect(self, other, minimize = True):
        new = FST()
        new.fsthandle = self._fomacallbinary(other, foma_fsm_intersect, minimize)
        return new

    def minus(self, other, minimize = True):
        new = FST()
        new.fsthandle = self._fomacallbinary(other, foma_fsm_minus, minimize)
        return new

    def concat(self, other, minimize = True):
        new = FST()
        new.fsthandle = self._fomacallbinary(other, foma_fsm_concat, minimize)
        return new

    def compose(self, other, minimize = True):
        new = FST()
        new.fsthandle = self._fomacallbinary(other, foma_fsm_compose, minimize)
        return new
    
    def lower(self, minimize = True):
        new = FST()
        new.fsthandle = self._fomacallunary(foma_fsm_lower, minimize)
        return new

    def upper(self, minimize = True):
        new = FST()
        new.fsthandle = self._fomacallunary(foma_fsm_upper, minimize)
        return new

    def flatten(self):
        new = FST()
        eps_sym = FST('□')
        new.fsthandle = foma_fsm_flatten(foma_fsm_copy(self.fsthandle), foma_fsm_copy(eps_sym.fsthandle))
        return new
    
class MTFSM(FST):

    def __init__(self, regex = False, numtapes = 2):
        if isinstance(regex, str) or isinstance(regex, unicode):
            FST.__init__(self, regex)
            eps_sym = FST('□')
            self.fsthandle = foma_fsm_flatten(foma_fsm_copy(self.fsthandle), foma_fsm_copy(eps_sym.fsthandle))
            self.numtapes = numtapes
        elif isinstance(regex, FST):
            self.fsthandle = foma_fsm_copy(regex.fsthandle)
            self.regex = None
            self.numtapes = numtapes
        else:
            self.fsthandle = None
            self.regex = None
        
        
    def __str__(self):
        if not self.fsthandle:
            raise ValueError('FSM not defined')
        foma_fsm_count(self.fsthandle)
        s  = 'Name: %s\n' % self.fsthandle.contents.name
        s += 'States: %i\n' % self.fsthandle.contents.statecount
        s += 'Transitions: %i\n' % self.fsthandle.contents.arccount
        s += 'Final states: %i\n' % self.fsthandle.contents.finalcount
        s += 'Deterministic: %i\n' % self.fsthandle.contents.is_deterministic
        s += 'Minimized: %i\n' % self.fsthandle.contents.is_minimized
        s += 'Numtapes: %i\n' % self.numtapes
        return s
    
    def parse(self, word):
        #[word/□ .o. [0:?^(numtapes-1) ?]*].l & Grammar ;
        m = self.numtapes
        regx = (u'[{' + word + u'}/□ .o. [0:?^' + str(m-1) + u' ?]*].l')
        reg = FST(regx)
        gr = FST()
        gr.fsthandle = foma_fsm_copy(self.fsthandle)
        res = MTFSM(reg.intersect(gr), numtapes = m)        
        return res
    
    def join(self, other):
        """Joins two multitape FSMs by composition. E.g.
            [ a d ]     [ c □ □ ]                  [ a d □ ]
        A = [ b e ] B = [ d f g ], and A.join(B) = [ b e □ ]
            [ c □ ]     [ e f g ]                  [ c □ □ ]
                                                   [ d f g ]
                                                   [ e f g ] """

        m = self.numtapes        
        n = other.numtapes
        pada = FST('[0:□^' + str(m) +' [0:?^' + str(n-1) + ' - 0:□^' + str(n-1) + '] | ?^' + str(m) + ' 0:?^' + str(n-1) + ']*')
        padb = FST('[[0:?^' + str(m-1) + ' - 0:□^' + str(m-1) + ' ] 0:□^' + str(n) + '| 0:?^' + str(m-1) + ' ?^' + str(n) + ']*')
        extenda = self.compose(pada).lower()
        extendb = other.compose(padb).lower()
        flt = FST('~[?^' + str(m+n-1) +'* [□^' + str(m) + ' ?^' + str(n-1) + ' [?^' + str(m-1) + ' □^' + str(n) + ' |[?^' + str(m-1) +' - □^' + str(m-1) + ' ] □ [?^' + str(n-1) + ' - □^' + str(n-1) + ' ]] | ?^' + str(m-1) + ' □^' + str(n) + ' [□^' + str(m) + ' ?^' + str(n-1) + ' |[?^' + str(m-1) + ' - □^' + str(m-1) + ' ] □ [?^' + str(n-1) + ' - □^' + str(n-1) + ' ]]] ?*]')
        res = extenda & extendb & flt
        result = MTFSM(res, m + n - 1)
        return result

    def _apply(self, applyf, word = None):
        if not self.fsthandle:
            raise ValueError('FST not defined')
        applyerhandle = foma_apply_init(self.fsthandle)
        output = applyf(c_void_p(applyerhandle))
        while True:
            if output == None:
                foma_apply_clear(c_void_p(applyerhandle))
                return
            else:
                yield output
            if word:
                output = applyf(c_void_p(applyerhandle), None)
            else:
                output = applyf(c_void_p(applyerhandle))

    def __iter__(self):
        return self._mtwords()

    def __add__(self, other):
        return self.join(other)

    def _fmt(self, word):
        cols = word
        colchunks = [map(lambda z: len(z), word[x:x+self.numtapes]) for x in xrange(0, len(word), self.numtapes)]
        col_widths = [max(x) for x in colchunks]
        format = '  '.join(['%%-%ds' % width for width in col_widths])
        # string to rows
        rows = [[word[y] for y in xrange(x, len(word), self.numtapes)] for x in xrange(self.numtapes)]
        s = ''
        for row in rows:
            #s += format % tuple(row) + '\n'
            s += ''.join(row) + '\n'
        return s

    def _mtwords(self):
        applyf = foma_apply_upper_words
        if not self.fsthandle:
            raise ValueError('FST not defined')
        applyerhandle = foma_apply_init(self.fsthandle)
        toksym = '\x07'
        foma_apply_set_space_symbol(c_void_p(applyerhandle), c_char_p(toksym))
        output = applyf(c_void_p(applyerhandle))
        while True:
            if output is None:
                foma_apply_clear(c_void_p(applyerhandle))
                return
            else:
                yield self._fmt(output[:-1].split('\x07'))
            output = applyf(c_void_p(applyerhandle))



#encoding: utf-8
import re, sys
from copy import copy
import codecs
from typing import Pattern

if (sys.version_info > (3, 0)):
	#python 3
	unicode = str

class UD_collection():
	"""docstring for UD_collection"""
	def __init__(self, file_handle):
		self.__iter_index = 0
		sentence = []
		self.sentences = []
		for line in file_handle:
			if len(line.replace(u"\n", "")) == 0:
				if len(sentence) > 0:
					self.sentences.append(parse_sentence(sentence))
				sentence = []
			else:
				sentence.append(line.replace("\n", ""))
		if len(sentence) > 0:
			#add last sentence
			self.sentences.append(parse_sentence(sentence))

	def find_sentences(self, query={}, head_query={}, match_range_tokens = False, match_empty_nodes = False, enhanced_dependencies=False, return_root= False):
		results = []
		for sentence in self.sentences:
			r = sentence.find(query, head_query, match_range_tokens, match_empty_nodes, enhanced_dependencies)
			if return_root and len(r) > 0:
				results.append(sentence)
			else:
				results.extend(r)
		return results
    
	def get_unique_feats(self, delimiter="|"):
		feats = []
		for sentence in self:
			fs = sentence.get_unique_feats(delimiter)
			for f in fs:
				if f not in feats:
					feats.append(f)
		return feats

	def get_unique_attributes(self, attribute):
		feats = []
		for sentence in self:
			fs = sentence.get_unique_attributes(attribute)
			for f in fs:
				if f not in feats:
					feats.append(f)
		return feats

	def __getitem__(self, i):
		return self.sentences[i]

	def __iter__(self):
		return self

	def __len__(self):
		return len(self.sentences)

	def next(self):
		if self.__iter_index >= len(self.sentences):
			self.__iter_index = 0
			raise StopIteration
		else:
			self.__iter_index += 1
			return self.sentences[self.__iter_index-1]

	def __next__(self):
		#python 3
		return self.next()

	def __repr__(self):
		representation = ""
		for sentence in self.sentences:
			representation += unicode(sentence) + "\n"
		return representation
		
class UD_sentence():
	def __init__(self):
		self.__iter_index = 0
		self.comments = ""
		self.id = "0"
		self.children = []
		self.secondary_children = []

	def set_root(self, root):
		self.root =root
		self.find = root.find

	def get_unique_feats(self, delimiter="|"):
		children = self.find()
		feats = []
		for child in children:
			fs = child.get_feats(delimiter)
			for f in fs:
				if f not in feats:
					feats.append(f)
		return feats

	def get_unique_attributes(self, attribute):
		children = self.find()
		feats = []
		for child in children:
			f = child.get_attribute(attribute)
			if f not in feats:
					feats.append(f)
		return feats

	def __repr__(self):
		children = self.find()
		children.sort()
		representation = self.comments
		for child in children:
			representation = representation + unicode(child) + u"\n"
		return representation
	
	def __getitem__(self, i):
		c = self.find()
		c.sort()
		return c[i]

	def __iter__(self):
		self.__iter_children = self.find()
		self.__iter_children.sort()
		return self

	def __len__(self):
		return len(self.find())

	def next(self):
		if self.__iter_index >= len(self.__iter_children):
			self.__iter_index = 0
			raise StopIteration
		else:
			self.__iter_index += 1
			return self.__iter_children[self.__iter_index-1]

	def __next__(self):
		#python 3
		return self.next()

class UD_relation():
	"""docstring for UD_relation"""
	def __init__(self, node, relation, head, primary=True):
		self.node = node
		self.head = head
		self.relation = relation
		if primary:
			if head is not None:
				head.children.append(self)
			node.head = self
		else:
			head.secondary_children.append(self)
			node.heads.append(self)
		self.primary = primary

	def __repr__(self):
		return self.head.id + ":" + self.relation

	def __eq__(self, other):
		return repr(self) == repr(other)

	def __lt__(self, other):
		if "." in self.node.id:
			main_index, second_index = self.node.id.split(".")
			if "." in other.node.id:
				main_index_o, second_index_o = other.node.id.split(".")
				if main_index_o == main_index:
					return int(second_index) < int(second_index_o)
			elif main_index == other.node.id:
				return False
		s_dash = False
		if "-" in self.node.id:
			s_dash = True
			s_id = (int(self.node.id.split("-")[0]))
		else:
			s_id = float(self.node.id)
		o_dash = False
		if "-" in other.node.id:
			o_dash = True
			o_id = (int(other.node.id.split("-")[0]))
		else:
			o_id = float(other.node.id)
		if o_id == s_id:
			if s_dash:
				return True
			return False
		else:
			return s_id < o_id
		

class UD_node():
	"""docstring for UD_node"""
	def __init__(self, id, form, lemma, upostag, xpostag, feats, misc):
		self.id = id
		self.form = form 
		self.lemma = lemma
		self.upostag = upostag
		self.pos = upostag
		self.xpostag = xpostag
		self.feats = feats
		self.misc = misc
		self.head = None
		self.children = []
		self.heads = []
		self.secondary_children = []

	def get_feats(self, delimiter="|"):
		return self.feats.split(delimiter)

	def get_attribute(self, attribute):
		if attribute == "id":
			return self.id
		elif attribute == "form":
			return self.form
		elif attribute == "lemma":
			return self.lemma
		elif attribute == "upostag" or attribute == "pos":
			return self.upostag
		elif attribute == "xpostag":
			return self.xpostag
		elif attribute == "feats":
			return self.feats
		elif attribute == "misc":
			return self.misc
		elif attribute == "deprel":
			return self.head.relation


	def find(self, query={}, head_query={}, match_range_tokens = False, match_empty_nodes = False, enhanced_dependencies=False):
		results = []
		passed = True
		if (not match_range_tokens and "-" in self.id) or (not match_empty_nodes and "." in self.id):
			passed = False
		else:
			for key in query:
				if key == "deprel":
					if self.head is None:
						attr = "root"
					else:
						attr = self.head.relation
				else:
					attr = self.get_attribute(key)
				v = query[key]
				if isinstance(v, Pattern):
					if v.match(attr) is None:
						passed = False
						break
				else:
					if v != attr:
						passed = False
						break
		
			head = self.head
			if head is not None and passed:
				if enhanced_dependencies:
					heads = self.heads + head
				else:
					heads = [head]
				head_pass = True
				for head in heads:
					head_pass = True
					for key in head_query:
						if key == "deprel":
							if self.head is None:
								attr = "root"
							else:
								attr = self.head.relation
						else:
							attr = getattr(self, key)
						v = head_query[key]
						if isinstance(v, Pattern):
							if v.match(attr) is None:
								head_pass = False
								break
						else:
							if v != attr:
								head_pass = False
								break
					if head_pass:
						break
				passed = head_pass
					

			elif head is None and len(head_query.keys()) >0:
				passed = False
		if passed:
			results.append(self)
		for child in self.children:
			r = child.node.find(query, head_query, match_range_tokens, match_empty_nodes, enhanced_dependencies)
			results.extend(r)
		return results


	def __repr__(self):
		head_repr = "0\troot"
		if self.head is not None:
			head_repr = self.head.head.id + "\t" + self.head.relation
		if len(self.heads) == 0:
			deps = "_"
		else:
			rels = copy(self.heads)
			rels = [x for x in rels if x.head != self]
			if self.head is not None:
				rels.append(self.head)
			rels.sort()
			deps = "|".join(str(x) for x in rels)
			if deps == "":
				deps = "_"
		return self.id + "\t" + self.form + "\t" + self.lemma + "\t" + self.upostag + "\t" + self.xpostag + "\t" + self.feats + "\t" + head_repr + "\t" + deps + "\t" + self.misc
	
	def __eq__(self, other):
		return repr(self) == repr(other)

	def __lt__(self, other):
		s_dash = False
		if "-" in self.id:
			s_dash = True
			s_id = (int(self.id.split("-")[0]))
		else:
			s_id = float(self.id)
		o_dash = False
		if "-" in other.id:
			o_dash = True
			o_id = (int(other.id.split("-")[0]))
		else:
			o_id = float(other.id)
		if o_id == s_id:
			if s_dash:
				return True
			return False
		else:
			return s_id < o_id

def parse_sentence(conll_u_sentence):
	if type(conll_u_sentence) == str or type(conll_u_sentence) == unicode:
		conll_u_sentence = conll_u_sentence.split("\n")
	nodes = {}
	relations = {}
	ud_sentence = UD_sentence()
	comments = ""
	for annotation in conll_u_sentence:
		if annotation == "":
			continue
		if annotation.startswith("#"):
			comments = comments + annotation + "\n"
			continue
		parts = annotation.split("\t")
		if "-" in parts[0]:
			#multi-part annotation --> skip for now
			continue
		node = UD_node(parts[0],parts[1],parts[2],parts[3],parts[4],parts[5],parts[9])
		nodes[parts[0]] = node
		relations[parts[0]] = [parts[6],parts[7],parts[8]]
	ud_sentence.comments = comments
	root = None
	nodes["0"] = ud_sentence
	for id in relations:
		relation = relations[id]
		head_id = relation[0]
		if head_id == "0":
			root = nodes[id]
		if "." in id and head_id == "_":
			head_id = id.split(".")[0]
		head_relation = UD_relation(nodes[id], relation[1], nodes[head_id])
		o_rel = relation[2]
		if o_rel == u"_":
			other_relations = []
		else:
			other_relations = o_rel.split("|")
		head_rel = head_id + ":" + relation[1]
		for other_relation in other_relations:
			if other_relation == head_rel:
				continue
			other_parts = other_relation.split(":")
			r = UD_relation(nodes[id], other_parts[1], nodes[other_parts[0]], False)
	ud_sentence.set_root(root)
	return ud_sentence

"""
root = parse_sentence(codecs.open("/Users/mikahama/Desktop/fi_test.conllu", "r", encoding="utf-8").read())
#print unicode(root)

nodes = root.find({"upostag":"NOUN", "deprel": "obl"})
for node in nodes:
	print unicode(node)
"""




import argparse
from . import semfi
from . import uralicApi

def download_semfi(langs):
	for lang in langs:
		semfi.download(lang)

def download(langs):
	for lang in langs:
		uralicApi.download(lang)

if __name__== "__main__":
	parser = argparse.ArgumentParser(description='UralicNLP download languages')
	parser.add_argument('-l','--languages', nargs='+', help='<Required> Languages to download', required=True)
	parser.add_argument("--semfi", action='store_true', help="Download semfi/semur for the specified languages")
	args = parser.parse_args()
	if args.semfi:
		download_semfi(args.languages)
	else:
		download(args.languages)




from uralicNLP.ud_tools import UD_collection
import urllib, requests
import sys

if (sys.version_info > (3, 0)):
	#python 3
	python_version = 3
else:
	python_version = 2

class LanguageNotSupported(Exception):
    pass

class BackendNotOnline(Exception):
    pass

def parse_text(sentence, language, **kwargs):
	if language != "fin":
		raise LanguageNotSupported("Language is not supported")
	return _turku_dependencies(sentence, **kwargs)


def _turku_dependencies(text, url="http://localhost:9876"):
	if python_version ==2 and type(text) is str:
		pass
	else:
		text = text.encode('utf-8')
	r = requests.post(url, data=text, headers={'content-type':'text/plain; charset=utf8'})
	if r.status_code != 200:
		raise BackendNotOnline("Couldn't connect to the docker server (code "+ str(r.status_code) +") at " + url + "\n Download the docker container on https://hub.docker.com/r/kazhar/finnish-dep-parser/ and make sure it is running.")
	conl = r.text
	sent = UD_collection(conl.split("\n"))
	return sent



from sklearn.cluster import AffinityPropagation, HDBSCAN
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity

class NotSupportedException(Exception):
	pass


def _get_matrix(embeddings):
	m = []
	embeddings = [np.array(e).reshape(1, -1) for e in embeddings]
	for em in embeddings:
		r = []
		for e in embeddings:
			r.append(cosine_similarity(em, e)[0][0])
		m.append(r)
	return m

def similarity(e1, e2):
	e1 = np.array(e1).reshape(1, -1)
	e2 = np.array(e2).reshape(1, -1)
	return cosine_similarity(e1, e2)[0][0]

def _centroid(vecs):
	return np.mean(vecs, axis=0)

def _semantic_clusters(embeddings, group=True, method="affinity", **kwargs):
	if method == "affinity":
		return _affinity_clusters(embeddings, group=group, **kwargs)
	elif method == "hdbscan":
		return _hdbscan_clusters(embeddings, group=group, **kwargs)
	else:
		raise NotSupportedException("Method must be affinity or hdbscan")


def _hdbscan_clusters(embeddings, group=True, min_cluster_size=2, **kwargs):
	hdb = HDBSCAN(min_cluster_size=min_cluster_size, **kwargs)
	u = hdb.fit([np.array(e) for e in embeddings])
	if group:
		return _group_results(hdb.labels_)
	else:
		return hdb.labels_

def _affinity_clusters(embeddings, group=True, **kwargs):
	m = np.array(_get_matrix(embeddings))
	agg = AffinityPropagation(affinity="precomputed", **kwargs)
	u = agg.fit_predict(m)
	if group:
		return _group_results(agg.labels_)
	else:
		return agg.labels_

def _group_results(labels):
	groups = []
	for x in range(len(set(labels))):
		groups.append([])
	for i, label in enumerate(labels):
		groups[label].append(i)
	return groups

def _cluster(embeddings, hierarchical_clustering, method="affinity", **kwargs):
	clusters = _semantic_clusters(embeddings, method=method, **kwargs)
	if not hierarchical_clustering:
		return clusters
	else:
		c_embeddings = embeddings
		c_map = clusters
		last_len = len(clusters)
		while True:
			c_embeddings = [_centroid([c_embeddings[v] for v in x]) for x in c_map]
			if len(c_embeddings) == 1:
				break
			x_map = _semantic_clusters(c_embeddings, False, method=method)
			c_map = _semantic_clusters(c_embeddings, True, method=method)
			groups = []
			for x in range(len(set(x_map))):
				groups.append([])
			for i, label in enumerate(x_map):
				groups[label].append(clusters[i])
			clusters = groups
			l = len(clusters)
			if l == last_len or l ==1:
				break
			else:
				last_len = l	
	return clusters


def _map_clusters(texts, clusters):
	for x in range(len(clusters)):
		c = clusters[x]
		if isinstance(c, list):
			_map_clusters(texts, c)
		else:
			clusters[x] = texts[c]

def cluster(texts, llm, return_ids=False, hierarchical_clustering=False, method="affinity", **kwargs):
	embeddings = [llm.embed(t) for t in texts]
	cs = _cluster(embeddings, hierarchical_clustering, method=method, **kwargs)
	if not return_ids:
		_map_clusters(texts, cs)
	return cs

def cluster_endangered(texts, llm, lang, dict_lang, return_ids=False, hierarchical_clustering=False, method="affinity", **kwargs):
	embeddings = [llm.embed_endangered(t, lang, dict_lang) for t in texts]
	cs = _cluster(embeddings, hierarchical_clustering, method=method, **kwargs)
	if not return_ids:
		_map_clusters(texts, cs)
	return cs



from tinydb import TinyDB, Query
import argparse
import os

try:
	from pymongo import MongoClient
except:
	pass

class EmptyDatabaseException(Exception):
    """Base class for exceptions in this module."""
    pass

class DictionaryInterface(object):
	"""docstring for DictionaryInterface"""
	def __init__(self, path, language):
		self.path = path
		self.language = language

	def find(self, word, lemmas):
		results = {"lemmatized":[]}
		results['exact_match'] = self._lemma_query(word)
		for lemma in lemmas:
			results["lemmatized"].append(self._lemma_query(lemma))
		results["other_languages"] = self._lang_query(word)
		return results

	def lemmas(self, group_by_pos=False):
		if group_by_pos:
			return self._list_lemmas_by_pos()
		else:
			return self._list_lemmas()

	def _list_lemmas(self):
		data = self._get_all()
		lems = []
		for l in data:
			try:
				lem = l["lg"]["l"]["#text"]
				lems.append(lem)
			except:
				pass
		return list(set(lems))

	def _list_lemmas_by_pos(self):
		data = self._get_all()
		lems = {}
		for l in data:
			try:
				lem = l["lg"]["l"]["#text"]
				pos = l["lg"]["l"]["@pos"]
				if pos not in lems:
					lems[pos] = []
				lems[pos].append(lem)
			except:
				pass
		for k in list(lems.keys()):
			lems[k] = list(set(lems[k]))
		return lems

	def _lemma_query(self, lemma):
		return []

	def _lang_query(self, lemma):
		return []

	def _get_all(self):
		return []

	def import_data(self):
		pass

class TinyDictionary(DictionaryInterface):
	"""docstring for TinyDictionary"""
	def __init__(self, path, language):
		if not os.path.isfile(path):
			raise EmptyDatabaseException("No dictionary.json available in " + path +"\nRunning uralicApi.download('"+language+"') may fix this if there is a dictionary available")
		self.db = TinyDB(path)

		if len(self.db) == 0:
			raise EmptyDatabaseException("The dictionary is empty for " + language + " in path " + path)
		super(TinyDictionary, self).__init__(path, language)

	def _get_all(self):
		return self.db.all()


	def _lemma_query(self, lemma):
		E = Query()
		res = self.db.search(E.lg.l["#text"] == lemma)
		return res

	def _lang_query(self, lemma):
		E = Query()
		T = Query()
		W = Query()
		res1 = self.db.search(E.mg.tg.any(T.t.any(W["#text"] == lemma)))
		res2 = self.db.search(E.mg.tg.any(T.t["#text"] == lemma))
		res3 = self.db.search(E.mg.tg.t["#text"] == lemma)

		return res1 + res2 +res3


class MongoDictionary(DictionaryInterface):
	"""docstring for MongoDictionary"""
	def __init__(self, path, language):
		client = MongoClient()
		self.db = client['uralicNLP_dicts']
		self.collection = self.db[language]
		self.empty = self.collection.count_documents({}) == 0
		super(MongoDictionary, self).__init__(path, language)

	def _get_all(self):
		self._check_empty()
		return self.collection.find()


	def _lemma_query(self, lemma):
		self._check_empty()
		res = self.collection.find({ "lg.l.#text" : lemma})
		return list(res)

	def _lang_query(self, lemma):
		self._check_empty()
		res = self.collection.find(
			{ "$or": [
		            { "mg.tg": { "$elemMatch": { "t": {"$elemMatch": {"#text": lemma }  } } } },
		            { "mg.tg.t.#text": lemma }
		        ]
		    }
		)

		return list(res)

	def _check_empty(self):
		if self.empty:
			raise EmptyDatabaseException("The dictionary database for '" + self.language + "' is empty.\nRun uralicApi.import_dictionary_to_db('" + self.language + "') first")


	def import_data(self):
		if not self.empty:
			self.collection.drop()
			self.collection = self.db[self.language]
		tiny = TinyDictionary(self.path, self.language)
		data = tiny._get_all()
		self.collection.insert_many(data)
		self.empty = self.collection.count_documents({}) == 0








#encoding: utf-8
import mikatools
import os
from .uralicApi import __find_writable_folder, __model_base_folders, ModelNotFound
import sqlite3
from collections import defaultdict

semfi_urls = "https://mikakalevi.com/whereis_semfi.json"

__connections = {}

def supported_languages():
	urls = mikatools.download_json(semfi_urls)
	return urls.keys()

def download(lang,show_progress=True):
	urls = mikatools.download_json(semfi_urls)
	if lang not in urls:
		raise "Language not supported. Currently supported languages " + ", ".join(urls.keys())
	download_to = os.path.join(__find_writable_folder(__model_base_folders()), lang)
	if not os.path.exists(download_to):
		os.makedirs(download_to)
	save_to = os.path.join(download_to, "sem.db")
	print("Downloading: "+ lang)
	mikatools.download_file(urls[lang], save_to, show_progress)

def __where_semfi(lang, safe=False):
	folders = __model_base_folders()
	for folder in folders:
		try_file = os.path.join(os.path.join(folder, lang), "sem.db")
		e = os.path.exists(try_file)
		if e:
			return try_file
	if safe:
		return None
	raise ModelNotFound()

def is_language_installed(language):
	path = __where_semfi(language, True)
	if path is None:
		return False
	return True

def __get_connection(lang):
	if lang in __connections:
		return __connections[lang]
	else:
		db_file = __where_semfi(lang)
		conn = sqlite3.connect(db_file)
		c = conn.cursor()
		__connections[lang] = c
		return c

def __column_headers(lang, table):
	if table == "words":
		hs = ["id", "word", "pos", "frequency", "relative_frequency"]
		if lang == "fin":
			hs.append("compund")
		else:
			hs.append("mwe")
	else:
		hs = ["word1", "word2", "relation", "frequency", "relative_frequency", "zscore"]
	return hs

def __add_titles(rows, lang, table):
	keys = __column_headers(lang, table)
	ret = []
	for row in rows:
		d = {}
		for i in range(len(row)):
			d[keys[i]] = row[i]
		ret.append(d)
	return ret

def __replace_by_word_object(word1, word2, relations, lang):
	c = __get_connection(lang)
	if word2 is not None:
		for relation in relations:
			relation["word1"] = word1
			relation["word2"] = word2
		return relations
	else:
		for relation in relations:
			relation["word1"] = word1
			relation["word2"] = get_word_by_id(relation["word2"], lang)
		return relations

def get_word(lemma, pos, lang):
	c = __get_connection(lang)
	c.execute('SELECT * FROM words WHERE word="'+lemma+'" and pos="' + pos +'"')
	all_rows = c.fetchall()
	rows = __add_titles(all_rows, lang, "words")
	if len(rows) > 0:
		return rows[0]
	else:
		return None

def get_word_by_id(id, lang):
	c = __get_connection(lang)
	c.execute('SELECT * FROM words WHERE id="'+id+'"')
	all_rows = c.fetchall()
	rows = __add_titles(all_rows, lang, "words")
	if len(rows) > 0:
		return rows[0]
	else:
		return None

def get_words(lemma, lang):
	c = __get_connection(lang)
	c.execute('SELECT * FROM words WHERE word="'+lemma+'"')
	all_rows = c.fetchall()
	return __add_titles(all_rows, lang, "words")

def get_all_relations(word_object, lang, sort=False, search_by_word2=False):
	sorting = ""
	if sort:
		sorting = " ORDER BY frequency DESC"
	c = __get_connection(lang)
	n = "1"
	if search_by_word2:
		n = "2"
	c.execute('SELECT * FROM relations WHERE word'+n+'="'+word_object["id"]+'"'+sorting)
	all_rows = c.fetchall()
	all_rows = __add_titles(all_rows, lang, "relations")
	rows = __replace_by_word_object(word_object, None, all_rows, lang)
	return rows

def get_by_relation(word_object, relation, lang, sort=False, search_by_word2=False):
	sorting = ""
	if sort:
		sorting = " ORDER BY frequency DESC"
	c = __get_connection(lang)
	n = "1"
	if search_by_word2:
		n = "2"
	c.execute('SELECT * FROM relations WHERE word'+n+'="'+word_object["id"]+'" and relation_name="' + relation +'"'+sorting)
	all_rows = c.fetchall()
	all_rows = __add_titles(all_rows, lang, "relations")
	rows = __replace_by_word_object(word_object, None, all_rows, lang)
	return rows

def get_by_word(word_object1, word_object2, lang, sort=False):
	sorting = ""
	if sort:
		sorting = " ORDER BY frequency DESC"
	c = __get_connection(lang)
	c.execute('SELECT * FROM relations WHERE word1="'+word_object1["id"]+'" and word2="' + word_object2["id"] +'"'+sorting)
	all_rows = c.fetchall()
	all_rows = __add_titles(all_rows, lang, "relations")
	rows = __replace_by_word_object(word_object1, word_object2, all_rows, lang)
	return rows

def get_by_word_and_relation(word_object1, word_object2, relation, lang, sort=False):
	sorting = ""
	if sort:
		sorting = " ORDER BY frequency DESC"
	c = __get_connection(lang)
	c.execute('SELECT * FROM relations WHERE word1="'+word_object1["id"]+'" and word2="' + word_object2["id"] +'" and relation_name="' + relation +'"'+sorting)
	all_rows = c.fetchall()
	all_rows = __add_titles(all_rows, lang, "relations")
	rows = __replace_by_word_object(word_object1, word_object2, all_rows, lang)
	return rows

def realtion_frequency(relations):
	rels = defaultdict(int)
	for relation in relations:
		rels[relation["relation"]] += relation["frequency"]
	return rels


def sort_by_frequency(lang, ascending=True, number=None, format=True):
	sort = "DESC"
	if ascending:
		sort = "ASC"
	c = __get_connection(lang)
	c.execute("SELECT * FROM words ORDER BY frequency " + sort)
	if number == None:
		all_rows = c.fetchall()
	else:
		all_rows = c.fetchmany(number)
	if format:
		return __add_titles(all_rows, lang, "words")
	else:
		return all_rows




#encoding: utf-8
from .uralicApi import analyze as uralic_api_analyze
from .uralicApi import __where_models as where_models
import os, sys
from subprocess import Popen, PIPE
from mikatools import open_write
import copy
import re

def _Cg3__parse_sentence(words, language, morphology_ignore_after=None, descriptive=True,remove_symbols=True, language_flags=False, words_analysis=None,neural_fallback=False):
	sentence = []
	if words_analysis is not None and len(words_analysis) < len(words):
		words_analysis = words_analysis + [[]]
	for i, word in enumerate(words):
		existing_analysis = None
		if words_analysis is not None:
			existing_analysis = words_analysis[i]
		analysis = __hfst_format(word, language, morphology_ignore_after,descriptive=descriptive, remove_symbols=remove_symbols, language_flags=language_flags, analysis=existing_analysis,neural_fallback=neural_fallback)
		sentence.extend(analysis)
	hfst_result_string = "\n".join(sentence)
	return hfst_result_string

def __hfst_format(word, language, morphology_ignore_after=None, descriptive=True,remove_symbols=True, language_flags=False, analysis=None,neural_fallback=False):
	if analysis is None:
		analysis = uralic_api_analyze(word, language,descriptive=descriptive,remove_symbols=remove_symbols, language_flags=language_flags,neural_fallback=neural_fallback)
	hfsts = []
	if len(analysis) == 0:
		hfsts.append(word + "\t" +word + "+?\tinf")
	for analys in analysis:
		if morphology_ignore_after is None:
			a = analys[0]
		else:
			a = analys[0].split(morphology_ignore_after)[0]
		hfsts.append(word + "\t" + a + "\t" + str(analys[1]))
	hfsts.append("")
	return hfsts

class Cg3():
	def __init__(self, language, morphology_languages=None):
		if morphology_languages is None:
			self.morphology_languages = language
		else:
			self.morphology_languages = morphology_languages
		model_path = where_models(language)
		cg_path = os.path.join(model_path, "cg")
		self.cg_path = cg_path
		self.language = language

	def disambiguate(self, words, morphology_ignore_after=None,descriptive=True,remove_symbols=True, temp_file=None, language_flags=False, morphologies=None, neural_fallback=False):
		hfst_output = __parse_sentence(words + [""], self.morphology_languages, morphology_ignore_after, descriptive=descriptive,remove_symbols=remove_symbols, language_flags=language_flags, words_analysis=morphologies, neural_fallback=neural_fallback)
		if temp_file is None:
			p1 = Popen(["echo", hfst_output], stdout=PIPE)
		else:
			f = open_write(temp_file)
			f.write(hfst_output)
			f.close()
			p1 = Popen(["cat", temp_file], stdout=PIPE)
		cg_conv = Popen(["cg-conv" ,"-f"], stdout=PIPE, stdin=p1.stdout)
		vislcg3 = Popen(['vislcg3', '--grammar', self.cg_path], stdout=PIPE, stdin=cg_conv.stdout)
		cg_results, error = vislcg3.communicate()
		return self.__parse_cg_results(cg_results)

	def __parse_cg_results(self, cg_results):
		if type(cg_results) is bytes:
			cg_results = cg_results.decode(sys.stdout.encoding)
		lines = cg_results.split("\n")
		results = []
		current_word = None
		current_list = []
		for line in lines:
			if line.startswith("\"<"):
				if current_word is not None:
					results.append((current_word, current_list))
				current_word = line[2:-2]
				current_list = []
			elif line.startswith("\t"):
				line = line[2:]
				parts = line.split("\" ", 1)
				if len(parts) < 2:
					continue
				w = Cg3Word(current_word, parts[0], parts[1].split(" "))
				current_list.append(w)
		return results
		
class Cg3Word():
	def __init__(self, form, lemma, morphology):
		self.form = form
		self.lemma = lemma
		self.morphology = morphology

	def __repr__(self):
		o = "<" + self.lemma + " - " + ", ".join(self.morphology) + ">"
		if type(o) is str:
			return o
		return o.encode("utf-8")

class Cg3Disambiguation():
	def __init__(self):
		self.arg = arg
		
class Cg3Pipe():
	def __init__(self, *args, **kwargs):
		self.cgs = args

	def _find_weight(self, cg_morphologies):
		for i, m in enumerate(cg_morphologies):
			if re.search(r"^\<W\:\d\.?\d+\>$",m):
				return i

	def _convert_morphologies(self, cg_results):
		analysis = []
		cg_results = copy.deepcopy(cg_results)
		for disambiguation in cg_results:
			possible_words = disambiguation[1]
			word_analysis = []
			for possible_word in possible_words:
				w = "".join(re.findall(r"\d+|\.", possible_word.morphology.pop(self._find_weight(possible_word.morphology))))
				w = float(w)
				r = [possible_word.lemma + "+" + "+".join(possible_word.morphology), w]
				word_analysis.append(r)
			analysis.append(word_analysis)
		return analysis

	def disambiguate(self, words, **kwargs):
		morphologies = None
		for cg in self.cgs:
			res = cg.disambiguate(words, morphologies=morphologies, **kwargs)
			morphologies = self._convert_morphologies(res)
		return res
		
		

